#include <iostream>
#include <fstream>
#include <vector>
#include <sstream>
#include <string>
#include <thread>
#include <chrono>
#include <algorithm>
#include <map>
#include <mutex>
#include <numeric>

struct State
{
  std::string StateID;
  char Symbol;
  char NewSymbol;
  bool Direction;
  std::string NextState;
};

std::mutex m;
void RunProgram(std::vector<State>& States, int& StartPos, std::string& Tape, bool& IsDone);
std::vector<State> ReadFromFile(std::string FileName, int& StartPos, std::string& Tape);

int main()
{
  /*Kintamieji, masyvai, vektoriai skirti tiuringo masinoms*/
  std::vector<std::thread> Threads;
  int NumberOfMachines;
  int StartPositions[64];
  bool IsDone[64]{false};
  std::string Tapes[64];
  std::vector<std::vector<State>> States;
  std::string FileName;

  /*Reikiamo turingo masinu skaiciaus ir failu pavadinimu surasymas*/
  std::cout << "Iveskite norima Tiuringo sim. skaiciu (1-64): ";
  std::cin >> NumberOfMachines;

  for (size_t i = 0; i < NumberOfMachines; i++)
  {
    std::cout << "Iveskite nuskaitomo failo pavadinima: ";
    std::cin >> FileName;

    auto State = ReadFromFile(FileName, StartPositions[i], Tapes[i]);
    States.push_back(State);
  }

  /*thread sumetami i Threads vektoriu*/
  for (size_t i = 0; i < NumberOfMachines; i++)
  {
    Threads.push_back(std::thread(RunProgram, std::ref(States[i]), std::ref(StartPositions[i]), std::ref(Tapes[i]), std::ref(IsDone[i])));
  }
  
  /*Kol visos masinos nepasieke pabaigos tol sukasi while*/
  bool ShouldRun = true;
  while (ShouldRun)
  {
    int Sum = std::accumulate(IsDone, IsDone + NumberOfMachines, 0);

    if (Sum == NumberOfMachines)
    {
      ShouldRun = false;
    }
    
    for (size_t i = 0; i < NumberOfMachines; i++)
    {
      std::cout << Tapes[i] << "\n";
      std::cout << "-----------------------------------------------------------------------------------------------" << "\n";
    }

    std::this_thread::sleep_for(std::chrono::milliseconds(100));
    system("cls");
  }

  /*Galutinis juostu atspausdinimas*/
  for (size_t i = 0; i < NumberOfMachines; i++)
  {
    std::cout << Tapes[i] << "\n";
    std::cout << "-----------------------------------------------------------------------------------------------" << "\n";
  }
  system("pause");

  for (auto& t : Threads)
  {
    if (t.joinable())
    {
      t.join();
    }
  }
}

//Failo nuskaitymas
std::vector<State> ReadFromFile(std::string FileName, int& StartPos, std::string& Tape)
{
  std::ifstream is(FileName);
  std::string Temp;
  std::vector<State> States;
  State TempState;

  is >> StartPos >> Tape;
  StartPos--;

  while (is.good())
  {
    std::getline(is, Temp);
    if (!Temp.empty())
    {
      std::istringstream ss(Temp);
      ss >> TempState.StateID >> TempState.Symbol >> TempState.NewSymbol;
      ss >> Temp;

      if (Temp == "R")
      {
        TempState.Direction = 1;
      }
      else
      {
        TempState.Direction = 0;
      }
      ss >> TempState.NextState;
      States.push_back(TempState);
    }
  }
  return States;
}

//Tiuringo masina
void RunProgram(std::vector<State>& States, int& StartPos, std::string& Tape, bool& IsDone)
{
  bool ShouldRun = true;
  int HeadPos = StartPos;
  std::string CurrentState = States[0].StateID;

  /*std::string Str = std::to_string(HeadPos) + "rez.txt";
  std::ofstream rez(Str);*/

  std::string T1;

  while (ShouldRun)
  {
    for (size_t i = 0; i < States.size(); i++)
    {
      T1 = Tape;
      if (States[i].Symbol == Tape[HeadPos] && States[i].StateID == CurrentState)
      {
        Tape[HeadPos] = States[i].NewSymbol;

        if (States[i].Direction)
        {
          if (HeadPos > Tape.size())
          {
            Tape.push_back(' ');
          }
          HeadPos++;
        }

        if (!States[i].Direction)
        {
          if (HeadPos - 1 < 0)
          {
            ShouldRun = false;
          }
          else
          {
            HeadPos--;
          }
        }

        CurrentState = States[i].NextState;
        i = 0;

        if (CurrentState == "X")
        {
          IsDone = true;
          ShouldRun = false;
        }
        /*m.lock();
        if (T1 != Tape)
        {
          rez << Tape << "\n";
        }
        m.unlock();*/
        std::this_thread::sleep_for(std::chrono::milliseconds(5));
        break;
      }
    }
  }
  IsDone = true;
 /* rez.close();*/
}